// matrix class (OpenGL RHS)
// -Kyoung Shin Park (2010 Spring)
//////////////////////////////////////////////////////////////////////
#include "matrix4x4.h"

namespace oglclass 
{
const float matrix4x4::EPSILON = 1e-6f;

matrix4x4::matrix4x4(float m0, float m4, float  m8, float m12,
                     float m1, float m5, float  m9, float m13,
                     float m2, float m6, float m10, float m14,
                     float m3, float m7, float m11, float m15)
{
    m_[0] = m0; m_[4] = m4; m_[8]  = m8;  m_[12] = m12;
    m_[1] = m1; m_[5] = m5; m_[9]  = m9;  m_[13] = m13;
    m_[2] = m2; m_[6] = m6; m_[10] = m10; m_[14] = m14;
    m_[3] = m3; m_[7] = m7; m_[11] = m11; m_[15] = m15;
}

matrix4x4::matrix4x4(const matrix4x4& other)
{
    m_[0] = other.m_[0]; m_[4] = other.m_[4]; m_[8]  = other.m_[8];  m_[12] = other.m_[12];
    m_[1] = other.m_[1]; m_[5] = other.m_[5]; m_[9]  = other.m_[9];  m_[13] = other.m_[13];
    m_[2] = other.m_[2]; m_[6] = other.m_[6]; m_[10] = other.m_[10]; m_[14] = other.m_[14];
    m_[3] = other.m_[3]; m_[7] = other.m_[7]; m_[11] = other.m_[11]; m_[15] = other.m_[15];
}

void matrix4x4::identity(void)
{
    m_[0] = 1.0f; m_[4] = 0.0f; m_[8]  = 0.0f; m_[12] = 0.0f;
    m_[1] = 0.0f; m_[5] = 1.0f; m_[9]  = 0.0f; m_[13] = 0.0f;
    m_[2] = 0.0f; m_[6] = 0.0f; m_[10] = 1.0f; m_[14] = 0.0f;
    m_[3] = 0.0f; m_[7] = 0.0f; m_[11] = 0.0f; m_[15] = 1.0f;
}

matrix4x4 matrix4x4::transpose(void)
{
	matrix4x4 result;

    result[0] = m_[0];  result[4] = m_[1];  result[8]  = m_[2];  result[12] = m_[3];
    result[1] = m_[4];  result[5] = m_[5];  result[9]  = m_[6];  result[13] = m_[7];
    result[2] = m_[8];	result[6] = m_[9];  result[10] = m_[10]; result[14] = m_[11];
    result[3] = m_[12]; result[7] = m_[13]; result[11] = m_[14]; result[15] = m_[15];

	return result;
}

float matrix4x4::determinant() const
{
    return (m_[0] * m_[5] - m_[1] * m_[4])
        * (m_[10] * m_[15] - m_[11] * m_[14])
        - (m_[0] * m_[6] - m_[2] * m_[4])
        * (m_[9] * m_[15] - m_[11] * m_[13])
        + (m_[0] * m_[7] - m_[3] * m_[4])
        * (m_[9] * m_[14] - m_[10] * m_[13])
        + (m_[1] * m_[6] - m_[2] * m_[5])
        * (m_[8] * m_[15] - m_[11] * m_[12])
        - (m_[1] * m_[7] - m_[3] * m_[5])
        * (m_[8] * m_[14] - m_[10] * m_[12])
        + (m_[2] * m_[7] - m_[3] * m_[6])
        * (m_[8] * m_[13] - m_[9] * m_[12]);
}

matrix4x4 matrix4x4::inverse(void)
{
    // This method of computing the inverse of a 4x4 matrix is based
    // on a similar function found in Paul Nettle's matrix template class
    // (http://www.fluidstudios.com).
    // If the inverse doesn't exist for this matrix, then the identity
    // matrix will be returned.

    matrix4x4 result;
    float det = determinant();

    if (matrix4x4::closeEnough(det, 0.0f)) {
        result.identity();
    }
    else {
        det = 1.0f / det;

        result.m_[0] = det * (m_[5] * (m_[10] * m_[15] - m_[11] * m_[14]) + m_[6] * (m_[11] * m_[13] - m_[9] * m_[15]) + m_[7] * (m_[9] * m_[14] - m_[10] * m_[13]));
        result.m_[1] = det * (m_[9] * (m_[2] * m_[15] - m_[3] * m_[14]) + m_[10] * (m_[3] * m_[13] - m_[1] * m_[15]) + m_[11] * (m_[1] * m_[14] - m_[2] * m_[13]));
        result.m_[2] = det * (m_[13] * (m_[2] * m_[7] - m_[3] * m_[6]) + m_[14] * (m_[3] * m_[5] - m_[1] * m_[7]) + m_[15] * (m_[1] * m_[6] - m_[2] * m_[5]));
        result.m_[3] = det * (m_[1] * (m_[7] * m_[10] - m_[6] * m_[11]) + m_[2] * (m_[5] * m_[11] - m_[7] * m_[9]) + m_[3] * (m_[6] * m_[9] - m_[5] * m_[10]));
        
        result.m_[4] = det * (m_[6] * (m_[8] * m_[15] - m_[11] * m_[12]) + m_[7] * (m_[10] * m_[12] - m_[8] * m_[14]) + m_[4] * (m_[11] * m_[14] - m_[10] * m_[15]));
        result.m_[5] = det * (m_[10] * (m_[0] * m_[15] - m_[3] * m_[12]) + m_[11] * (m_[2] * m_[12] - m_[0] * m_[14]) + m_[8] * (m_[3] * m_[14] - m_[2] * m_[15]));
        result.m_[6] = det * (m_[14] * (m_[0] * m_[7] - m_[3] * m_[4]) + m_[15] * (m_[2] * m_[4] - m_[0] * m_[6]) + m_[12] * (m_[3] * m_[6] - m_[2] * m_[7]));
        result.m_[7] = det * (m_[2] * (m_[7] * m_[8] - m_[4] * m_[11]) + m_[3] * (m_[4] * m_[10] - m_[6] * m_[8]) + m_[0] * (m_[6] * m_[11] - m_[7] * m_[10]));
        
        result.m_[8] = det * (m_[7] * (m_[8] * m_[13] - m_[9] * m_[12]) + m_[4] * (m_[9] * m_[15] - m_[11] * m_[13]) + m_[5] * (m_[11] * m_[12] - m_[8] * m_[15]));
        result.m_[9] = det * (m_[11] * (m_[0] * m_[13] - m_[1] * m_[12]) + m_[8] * (m_[1] * m_[15] - m_[3] * m_[13]) + m_[9] * (m_[3] * m_[12] - m_[0] * m_[15]));
        result.m_[10] = det * (m_[15] * (m_[0] * m_[5] - m_[1] * m_[4]) + m_[12] * (m_[1] * m_[7] - m_[3] * m_[5]) + m_[13] * (m_[3] * m_[4] - m_[0] * m_[7]));
        result.m_[11] = det * (m_[3] * (m_[5] * m_[8] - m_[4] * m_[9]) + m_[0] * (m_[7] * m_[9] - m_[5] * m_[11]) + m_[1] * (m_[4] * m_[11] - m_[7] * m_[8]));
        
        result.m_[12] = det * (m_[4] * (m_[10] * m_[13] - m_[9] * m_[14]) + m_[5] * (m_[8] * m_[14] - m_[10] * m_[12]) + m_[6] * (m_[9] * m_[12] - m_[8] * m_[13]));
        result.m_[13] = det * (m_[8] * (m_[2] * m_[13] - m_[1] * m_[14]) + m_[9] * (m_[0] * m_[14] - m_[2] * m_[12]) + m_[10] * (m_[1] * m_[12] - m_[0] * m_[13]));
        result.m_[14] = det * (m_[12] * (m_[2] * m_[5] - m_[1] * m_[6]) + m_[13] * (m_[0] * m_[6] - m_[2] * m_[4]) + m_[14] * (m_[1] * m_[4] - m_[0] * m_[5]));
        result.m_[15] = det * (m_[0] * (m_[5] * m_[10] - m_[6] * m_[9]) + m_[1] * (m_[6] * m_[8] - m_[4] * m_[10]) + m_[2] * (m_[4] * m_[9] - m_[5] * m_[8]));
    }

    return result;
}

void matrix4x4::translate(vector3& trans)
{
    identity();
    m_[12] = trans[0];
    m_[13] = trans[1];
    m_[14] = trans[2];
}

void matrix4x4::translate(float x, float y, float z)
{
    identity();
    m_[12] = x;
    m_[13] = y;
    m_[14] = z;
}

// rotation using arbitrary axis and angle
void matrix4x4::rotate(const float& angle, vector3& axis)
{
    float s = sinf(DEGTORAD(angle));
    float c = cosf(DEGTORAD(angle));

    axis.normalize();

    float ux = axis[0];
    float uy = axis[1];
    float uz = axis[2];
    
    m_[0]  = c + (1-c) * pow(ux,2);
    m_[1]  = (1-c) * ux*uy + s*uz;
    m_[2]  = (1-c) * ux*uz - s*uy;
    m_[3]  = 0;
    
    m_[4]  = (1-c) * uy*ux - s*uz;
    m_[5]  = c + (1-c) * pow(uy,2);
    m_[6]  = (1-c) * uy*uz + s*ux;
    m_[7]  = 0;
    
    m_[8]  = (1-c) * uz*ux + s*uy;
    m_[9]  = (1-c) * uz*uy - s*ux;
    m_[10] = c + (1-c) * pow(uz,2);
    m_[11] = 0;
    
    m_[12] = 0;
    m_[13] = 0;
    m_[14] = 0;
    m_[15] = 1;
}

void matrix4x4::rotate(const float &angle, unsigned char axis)
{
    float s = sinf(DEGTORAD(angle));
    float c = cosf(DEGTORAD(angle));

    identity();

	if (axis == 'x')
	{
		m_[5]  =  c;
		m_[6]  =  s;
		m_[9]  = -s;
		m_[10] =  c;
	}
	else if (axis == 'y')
	{
		m_[0]  =  c;
		m_[2]  = -s;
		m_[8]  =  s;
		m_[10] =  c;
	}
	else if (axis == 'z')
	{
		m_[0] =  c;
		m_[1] =  s;
		m_[4] = -s;
		m_[5] =  c;
	}
}

// rotation using arbitrary axis and angle
void matrix4x4::rotate(float angle, float x, float y, float z)
{
    float s = sinf(DEGTORAD(angle));
    float c = cosf(DEGTORAD(angle));

	vector3 axis;
	axis.set(x, y, z);
    axis.normalize();

    float ux = axis[0];
    float uy = axis[1];
    float uz = axis[2];
    
    m_[0]  = c + (1-c) * pow(ux,2);
    m_[1]  = (1-c) * ux*uy + s*uz;
    m_[2]  = (1-c) * ux*uz - s*uy;
    m_[3]  = 0;
    
    m_[4]  = (1-c) * uy*ux - s*uz;
    m_[5]  = c + (1-c) * pow(uy,2);
    m_[6]  = (1-c) * uy*uz + s*ux;
    m_[7]  = 0;
    
    m_[8]  = (1-c) * uz*ux + s*uy;
    m_[9]  = (1-c) * uz*uy - s*ux;
    m_[10] = c + (1-c) * pow(uz,2);
    m_[11] = 0;
    
    m_[12] = 0;
    m_[13] = 0;
    m_[14] = 0;
    m_[15] = 1;
}

void matrix4x4::scale(vector3& scale)
{
    identity();
    m_[0]  = scale[0];
    m_[5]  = scale[1];
    m_[10] = scale[2];
}

void matrix4x4::scale(float x, float y, float z)
{
    identity();
    m_[0]  = x;
    m_[5]  = y;
    m_[10] = z;
}

void matrix4x4::transformPoint(vector3 *vec)
{
    vector3 &v = *vec;

	float x = v[0];
	float y = v[1];
	float z = v[2];

    v[0] = x * m_[0] + y * m_[4] + z * m_[8]  + m_[12];
    v[1] = x * m_[1] + y * m_[5] + z * m_[9]  + m_[13];
    v[2] = x * m_[2] + y * m_[6] + z * m_[10] + m_[14];
}

void matrix4x4::transformVector(vector3 *vec)
{
    vector3 &v = *vec;
	float x = v[0];
	float y = v[1];
	float z = v[2];

    v[0] = x * m_[0] + y * m_[4] + z * m_[8];
    v[1] = x * m_[1] + y * m_[5] + z * m_[9];
    v[2] = x * m_[2] + y * m_[6] + z * m_[10];
}

matrix4x4 matrix4x4::operator+ (const matrix4x4 &other)
{
    matrix4x4 result;

    result.m_[0]  = m_[0]  + other.m_[0];
    result.m_[1]  = m_[1]  + other.m_[1];
    result.m_[2]  = m_[2]  + other.m_[2];
    result.m_[3]  = m_[3]  + other.m_[3];

    result.m_[4]  = m_[4]  + other.m_[4];
    result.m_[5]  = m_[5]  + other.m_[5];
    result.m_[6]  = m_[6]  + other.m_[6];
    result.m_[7]  = m_[7]  + other.m_[7];

    result.m_[8]  = m_[8]  + other.m_[8];
    result.m_[9]  = m_[9]  + other.m_[9];
    result.m_[10] = m_[10] + other.m_[10];
    result.m_[11] = m_[11] + other.m_[11];

    result.m_[12] = m_[12] + other.m_[12];
    result.m_[13] = m_[13] + other.m_[13];
    result.m_[14] = m_[14] + other.m_[14];
    result.m_[15] = m_[15] + other.m_[15];

    return result;
}

matrix4x4 matrix4x4::operator- (const matrix4x4 &other)
{
    matrix4x4 result;

    result.m_[0]  = m_[0]  - other.m_[0];
    result.m_[1]  = m_[1]  - other.m_[1];
    result.m_[2]  = m_[2]  - other.m_[2];
    result.m_[3]  = m_[3]  - other.m_[3];

    result.m_[4]  = m_[4]  - other.m_[4];
    result.m_[5]  = m_[5]  - other.m_[5];
    result.m_[6]  = m_[6]  - other.m_[6];
    result.m_[7]  = m_[7]  - other.m_[7];

    result.m_[8]  = m_[8]  - other.m_[8];
    result.m_[9]  = m_[9]  - other.m_[9];
    result.m_[10] = m_[10] - other.m_[10];
    result.m_[11] = m_[11] - other.m_[11];

    result.m_[12] = m_[12] - other.m_[12];
    result.m_[13] = m_[13] - other.m_[13];
    result.m_[14] = m_[14] - other.m_[14];
    result.m_[15] = m_[15] - other.m_[15];

    return result;
}

matrix4x4 matrix4x4::operator* (const matrix4x4 &other)
{
    matrix4x4 result;

    result.m_[0]  = (m_[0]*other.m_[0])  + (m_[4]*other.m_[1])  + (m_[8]*other.m_[2])   + (m_[12]*other.m_[3]);
    result.m_[1]  = (m_[1]*other.m_[0])  + (m_[5]*other.m_[1])  + (m_[9]*other.m_[2])   + (m_[13]*other.m_[3]);
    result.m_[2]  = (m_[2]*other.m_[0])  + (m_[6]*other.m_[1])  + (m_[10]*other.m_[2])  + (m_[14]*other.m_[3]);
    result.m_[3]  = (m_[3]*other.m_[0])  + (m_[7]*other.m_[1])  + (m_[11]*other.m_[2])  + (m_[15]*other.m_[3]);

    result.m_[4]  = (m_[0]*other.m_[4])  + (m_[4]*other.m_[5])  + (m_[8]*other.m_[6])   + (m_[12]*other.m_[7]);
    result.m_[5]  = (m_[1]*other.m_[4])  + (m_[5]*other.m_[5])  + (m_[9]*other.m_[6])   + (m_[13]*other.m_[7]);
    result.m_[6]  = (m_[2]*other.m_[4])  + (m_[6]*other.m_[5])  + (m_[10]*other.m_[6])  + (m_[14]*other.m_[7]);
    result.m_[7]  = (m_[3]*other.m_[4])  + (m_[7]*other.m_[5])  + (m_[11]*other.m_[6])  + (m_[15]*other.m_[7]);

    result.m_[8]  = (m_[0]*other.m_[8])  + (m_[4]*other.m_[9])  + (m_[8]*other.m_[10])  + (m_[12]*other.m_[11]);
    result.m_[9]  = (m_[1]*other.m_[8])  + (m_[5]*other.m_[9])  + (m_[9]*other.m_[10])  + (m_[13]*other.m_[11]);
    result.m_[10] = (m_[2]*other.m_[8])  + (m_[6]*other.m_[9])  + (m_[10]*other.m_[10]) + (m_[14]*other.m_[11]);
    result.m_[11] = (m_[3]*other.m_[8])  + (m_[7]*other.m_[9])  + (m_[11]*other.m_[10]) + (m_[15]*other.m_[11]);

    result.m_[12] = (m_[0]*other.m_[12]) + (m_[4]*other.m_[13]) + (m_[8]*other.m_[14])  + (m_[12]*other.m_[15]);
    result.m_[13] = (m_[1]*other.m_[12]) + (m_[5]*other.m_[13]) + (m_[9]*other.m_[14])  + (m_[13]*other.m_[15]);
    result.m_[14] = (m_[2]*other.m_[12]) + (m_[6]*other.m_[13]) + (m_[10]*other.m_[14]) + (m_[14]*other.m_[15]);
    result.m_[15] = (m_[3]*other.m_[12]) + (m_[7]*other.m_[13]) + (m_[11]*other.m_[14]) + (m_[15]*other.m_[15]);

    return result;
}

matrix4x4 matrix4x4::operator* (const float scalar)
{
    matrix4x4 result;

    result.m_[0]  = m_[0]  * scalar;
    result.m_[1]  = m_[1]  * scalar;
    result.m_[2]  = m_[2]  * scalar;
    result.m_[3]  = m_[3]  * scalar;

    result.m_[4]  = m_[4]  * scalar;
    result.m_[5]  = m_[5]  * scalar;
    result.m_[6]  = m_[6]  * scalar;
    result.m_[7]  = m_[7]  * scalar;

    result.m_[8]  = m_[8]  * scalar;
    result.m_[9]  = m_[9]  * scalar;
    result.m_[10] = m_[10] * scalar;
    result.m_[11] = m_[11] * scalar;

    result.m_[12] = m_[12] * scalar;
    result.m_[13] = m_[13] * scalar;
    result.m_[14] = m_[14] * scalar;
    result.m_[15] = m_[15] * scalar;

    return result;
}

float& matrix4x4::operator[] (const int i)
{
	if (i > 0 && i <= 16)
		return m_[i];
	return m_[0];
}

float matrix4x4::operator[] (const int i) const
{
	if (i > 0 && i <= 16)
		return m_[i];
	return m_[0];
}

matrix4x4::operator float* () const
{
	return (float*)m_;
}


} // namespace

